#![allow(non_snake_case)]

extern crate serenity;

extern crate parking_lot;
extern crate serde_json;
extern crate typemap;

mod config_handler;
mod player;

use serenity::{
    builder::CreateEmbed,
    framework::standard::{
        macros::{group, help},
        Args, CommandGroup, CommandResult, HelpOptions, StandardFramework,
    },
    model::{channel::Message, id::UserId},
};

// This imports `typemap`'s `Key` as `TypeMapKey`.
use anyhow::Result;
use serenity::prelude::*;

use std::sync::{Arc, Mutex};

use config_handler::{ConfigHandler, Value};
use player::prelude::*;
use std::collections::HashSet;

create_settings_section!(
    Config,
    "Meta",
    {
        token: String,
        prefix: String,
        volume: f32,
    },
);

#[group]
#[commands(ip, list, pause, play, remove, skip, stop, tag)]
struct General;

struct HelpGroup {
    name: String,
    infos: Vec<HelpInfo>,
}

impl HelpGroup {
    fn new(name: &str) -> Self {
        Self {
            name: name.to_string(),
            infos: Vec::new(),
        }
    }

    fn add_info(mut self, info: HelpInfo) -> Self {
        self.infos.push(info);

        self
    }
}

struct HelpInfo {
    command: String,
    sub_commands: Vec<HelpSubInfo>,
}

impl HelpInfo {
    fn new(command: &str) -> Self {
        Self {
            command: command.to_string(),
            sub_commands: Vec::new(),
        }
    }

    fn add_sub(mut self, sub_command: impl Into<HelpSubInfo>) -> Self {
        self.sub_commands.push(sub_command.into());

        self
    }
}

enum HelpSubInfo {
    Empty,
    SubCommand(String),
}

impl<'a> From<&'a str> for HelpSubInfo {
    fn from(s: &'a str) -> Self {
        Self::SubCommand(s.to_string())
    }
}

impl From<String> for HelpSubInfo {
    fn from(s: String) -> Self {
        Self::SubCommand(s)
    }
}

impl From<Option<String>> for HelpSubInfo {
    fn from(o: Option<String>) -> Self {
        match o {
            Some(s) => Self::SubCommand(s.to_string()),
            None => Self::Empty,
        }
    }
}

fn create_help(embed: &mut CreateEmbed, groups: Vec<HelpGroup>) {
    for group in groups {
        let mut s = String::new();

        for info in group.infos {
            s += &format!("`{}`\n", info.command);

            for sub in info.sub_commands {
                s += &format!(
                    "> {}\n",
                    match sub {
                        HelpSubInfo::Empty => "<empty>".to_string(),
                        HelpSubInfo::SubCommand(sub_command) => sub_command,
                    }
                )
            }
        }

        embed.field(group.name, s, false);
    }
}

#[help]
fn my_help(
    context: &mut Context,
    msg: &Message,
    _args: Args,
    help_options: &'static HelpOptions,
    _groups: &[&'static CommandGroup],
    _owners: HashSet<UserId>,
) -> CommandResult {
    msg.channel_id.send_message(&context.http, |m| {
        m.embed(|embed| {
            embed.colour(help_options.embed_success_colour);

            let groups = vec![
                HelpGroup::new("General")
                    .add_info(HelpInfo::new("ip"))
                    .add_info(HelpInfo::new("help")),
                HelpGroup::new("Audio Control")
                    .add_info(
                        HelpInfo::new("play")
                            .add_sub(None)
                            .add_sub("--local")
                            .add_sub("--tag <tag name>")
                            .add_sub("<link>")
                            .add_sub("<name>"),
                    )
                    .add_info(HelpInfo::new("pause"))
                    .add_info(HelpInfo::new("stop"))
                    .add_info(HelpInfo::new("skip")),
                HelpGroup::new("Data Manipulation")
                    .add_info(HelpInfo::new("list").add_sub(None).add_sub("--tags"))
                    .add_info(HelpInfo::new("remove"))
                    .add_info(HelpInfo::new("tag").add_sub("<tag name>")),
            ];

            create_help(embed, groups);

            embed
        });
        m
    })?;

    Ok(())
}

fn main() -> Result<()> {
    // read config file
    let config_file = ConfigHandler::read_config("bot.conf")?;

    let config = Config::load(&config_file)?;

    let mut client = Client::new(&config.token, Handler).expect("Err creating client");

    let media_data = Arc::new(Mutex::new(
        MediaData::new("Penis1", Arc::clone(&client.voice_manager), config.volume)
            .expect("failed to create media data handle"),
    ));

    // Obtain a lock to the data owned by the client, and insert the client's
    // voice manager into it. This allows the voice manager to be accessible by
    // event handlers and framework commands.
    {
        let mut data = client.data.write();
        data.insert::<Media>(media_data.clone());
    }

    // We will fetch your bot's owners and id
    let (owners, bot_id) = match client.cache_and_http.http.get_current_application_info() {
        Ok(info) => {
            let mut owners = HashSet::new();
            owners.insert(info.owner.id);

            (owners, info.id)
        }
        Err(why) => panic!("Could not access application info: {:?}", why),
    };

    client.with_framework(
        // Configures the client, allowing for options to mutate how the
        // framework functions.
        //
        // Refer to the documentation for
        // `serenity::ext::framework::Configuration` for all available
        // configurations.
        StandardFramework::new()
            .configure(|c| {
                c.with_whitespace(true)
                    .on_mention(Some(bot_id))
                    .prefix(&config.prefix)
                    .owners(owners)
            })
            .group(&GENERAL_GROUP)
            .help(&MY_HELP),
    );

    let _ = client
        .start()
        .map_err(|why| println!("Client ended: {:?}", why));

    Ok(())
}
